From 8b5748ce5376f04cbfe52e2e57ba7a03f0acf5ed Mon Sep 17 00:00:00 2001
From: Naohiro KORIYAMA <nkoriyama@gmail.com>
Date: Mon, 19 Sep 2011 08:42:45 +0900
Subject: [PATCH] TS demux: #2985 Enhance TS demuxer to support duration and
 current time

---
 modules/demux/ts.c |  505 ++++++++++++++++++++++++++++++++++++++++++----------
 1 files changed, 414 insertions(+), 91 deletions(-)

diff --git a/modules/demux/ts.c b/modules/demux/ts.c
index 441fb9b..91afd5e 100644
--- a/modules/demux/ts.c
+++ b/modules/demux/ts.c
@@ -336,6 +336,12 @@ struct demux_sys_t
     /* how many TS packet we read at once */
     int         i_ts_read;
 
+    /* to determine length and time */
+    int         i_pid_first_pcr;
+    mtime_t     i_first_pcr;
+    mtime_t     i_current_pcr;
+    mtime_t     i_last_pcr;
+
     /* All pid */
     ts_pid_t    pid[8192];
 
@@ -401,6 +407,12 @@ static inline int PIDGet( block_t *p )
 
 static bool GatherPES( demux_t *p_demux, ts_pid_t *pid, block_t *p_bk );
 
+static block_t* ReadTSPacket( demux_t *p_demux );
+static mtime_t GetPCR( block_t *p_pkt );
+static int SeekToPCR( demux_t *p_demux, int64_t i_pos );
+static int Seek( demux_t *p_demux, double f_percent );
+static void GetFirstPCR( demux_t *p_demux );
+static void GetLastPCR( demux_t *p_demux );
 static void PCRHandle( demux_t *p_demux, ts_pid_t *, block_t * );
 
 static iod_descriptor_t *IODNew( int , uint8_t * );
@@ -788,6 +800,14 @@ static int Open( vlc_object_t *p_this )
     p_sys->b_silent = var_CreateGetBool( p_demux, "ts-silent" );
     p_sys->b_split_es = var_InheritBool( p_demux, "ts-split-es" );
 
+    bool can_seek = false;
+    stream_Control( p_demux->s, STREAM_CAN_SEEK, &can_seek );
+    if( can_seek )
+    {
+        GetFirstPCR( p_demux );
+        GetLastPCR( p_demux );
+    }
+
     while( !p_sys->b_file_out && p_sys->i_pmt_es <= 0 &&
            vlc_object_alive( p_demux ) )
     {
@@ -1026,59 +1046,11 @@ static int Demux( demux_t *p_demux )
     {
         bool         b_frame = false;
         block_t     *p_pkt;
-
-        /* Get a new TS packet */
-        if( !( p_pkt = stream_Block( p_demux->s, p_sys->i_packet_size ) ) )
+        if( !(p_pkt = ReadTSPacket( p_demux )) )
         {
-            msg_Dbg( p_demux, "eof ?" );
             return 0;
         }
 
-        /* Check sync byte and re-sync if needed */
-        if( p_pkt->p_buffer[0] != 0x47 )
-        {
-            msg_Warn( p_demux, "lost synchro" );
-            block_Release( p_pkt );
-
-            while( vlc_object_alive (p_demux) )
-            {
-                const uint8_t *p_peek;
-                int i_peek, i_skip = 0;
-
-                i_peek = stream_Peek( p_demux->s, &p_peek,
-                                      p_sys->i_packet_size * 10 );
-                if( i_peek < p_sys->i_packet_size + 1 )
-                {
-                    msg_Dbg( p_demux, "eof ?" );
-                    return 0;
-                }
-
-                while( i_skip < i_peek - p_sys->i_packet_size )
-                {
-                    if( p_peek[i_skip] == 0x47 &&
-                        p_peek[i_skip + p_sys->i_packet_size] == 0x47 )
-                    {
-                        break;
-                    }
-                    i_skip++;
-                }
-
-                msg_Dbg( p_demux, "skipping %d bytes of garbage", i_skip );
-                stream_Read( p_demux->s, NULL, i_skip );
-
-                if( i_skip < i_peek - p_sys->i_packet_size )
-                {
-                    break;
-                }
-            }
-
-            if( !( p_pkt = stream_Block( p_demux->s, p_sys->i_packet_size ) ) )
-            {
-                msg_Dbg( p_demux, "eof ?" );
-                return 0;
-            }
-        }
-
         if( p_sys->b_start_record )
         {
             /* Enable recording once synchronized */
@@ -1196,13 +1168,10 @@ static int Control( demux_t *p_demux, int i_query, va_list args )
     {
     case DEMUX_GET_POSITION:
         pf = (double*) va_arg( args, double* );
-        i64 = stream_Size( p_demux->s );
-        if( i64 > 0 )
-        {
-            double f_current = stream_Tell( p_demux->s );
-            *pf = f_current / (double)i64;
-        }
-        else
+
+        if( (p_sys->b_dvb_meta && p_sys->b_access_control) ||
+            p_sys->i_current_pcr - p_sys->i_first_pcr < 0 ||
+            p_sys->i_last_pcr - p_sys->i_first_pcr <= 0 )
         {
             int64_t i_time, i_length;
             if( !DVBEventInformation( p_demux, &i_time, &i_length ) && i_length > 0 )
@@ -1210,50 +1179,66 @@ static int Control( demux_t *p_demux, int i_query, va_list args )
             else
                 *pf = 0.0;
         }
+        else
+        {
+            int64_t i_time, i_length;
+            i_time = (p_sys->i_current_pcr - p_sys->i_first_pcr) * 100 / 9;
+            i_length = (p_sys->i_last_pcr - p_sys->i_first_pcr) * 100 / 9;
+            *pf = (double)i_time/(double)i_length;
+        }
         return VLC_SUCCESS;
+
     case DEMUX_SET_POSITION:
         f = (double) va_arg( args, double );
-        i64 = stream_Size( p_demux->s );
 
-        if( stream_Seek( p_demux->s, (int64_t)(i64 * f) ) )
-            return VLC_EGENERIC;
+        if( (p_sys->b_dvb_meta && p_sys->b_access_control) ||
+            p_sys->i_current_pcr - p_sys->i_first_pcr  < 0 ||
+            p_sys->i_last_pcr - p_sys->i_first_pcr <= 0 )
+        {
+            i64 = stream_Size( p_demux->s );
+            if( stream_Seek( p_demux->s, (int64_t)(i64 * f) ) )
+                return VLC_EGENERIC;
+        }
+        else
+        {
+            if( Seek( p_demux, f ) )
+                return VLC_EGENERIC;
+        }
 
         return VLC_SUCCESS;
-#if 0
 
     case DEMUX_GET_TIME:
         pi64 = (int64_t*)va_arg( args, int64_t * );
-        if( p_sys->i_time < 0 )
+        if( (p_sys->b_dvb_meta && p_sys->b_access_control) ||
+            p_sys->i_current_pcr - p_sys->i_first_pcr < 0 )
         {
-            *pi64 = 0;
-            return VLC_EGENERIC;
+            if( DVBEventInformation( p_demux, pi64, NULL ) )
+            {
+                *pi64 = 0;
+            }
+        }
+        else
+        {
+            *pi64 = (p_sys->i_current_pcr - p_sys->i_first_pcr) * 100 / 9;
         }
-        *pi64 = p_sys->i_time;
         return VLC_SUCCESS;
 
     case DEMUX_GET_LENGTH:
         pi64 = (int64_t*)va_arg( args, int64_t * );
-        if( p_sys->i_mux_rate > 0 )
+        if( (p_sys->b_dvb_meta && p_sys->b_access_control) ||
+            p_sys->i_last_pcr - p_sys->i_first_pcr <= 0 )
         {
-            *pi64 = INT64_C(1000000) * ( stream_Size( p_demux->s ) / 50 ) /
-                    p_sys->i_mux_rate;
-            return VLC_SUCCESS;
+            if( DVBEventInformation( p_demux, NULL, pi64 ) )
+            {
+                *pi64 = 0;
+            }
+        }
+        else
+        {
+            *pi64 = (p_sys->i_last_pcr - p_sys->i_first_pcr) * 100 / 9;
         }
-        *pi64 = 0;
-        return VLC_EGENERIC;
-#else
-    case DEMUX_GET_TIME:
-        pi64 = (int64_t*)va_arg( args, int64_t * );
-        if( DVBEventInformation( p_demux, pi64, NULL ) )
-            *pi64 = 0;
         return VLC_SUCCESS;
 
-    case DEMUX_GET_LENGTH:
-        pi64 = (int64_t*)va_arg( args, int64_t * );
-        if( DVBEventInformation( p_demux, NULL, pi64 ) )
-            *pi64 = 0;
-        return VLC_SUCCESS;
-#endif
     case DEMUX_SET_GROUP:
     {
         vlc_list_t *p_list;
@@ -1847,24 +1832,362 @@ static void ParsePES( demux_t *p_demux, ts_pid_t *pid )
     }
 }
 
-static void PCRHandle( demux_t *p_demux, ts_pid_t *pid, block_t *p_bk )
+static block_t* ReadTSPacket( demux_t *p_demux )
 {
-    demux_sys_t   *p_sys = p_demux->p_sys;
-    const uint8_t *p = p_bk->p_buffer;
+    demux_sys_t *p_sys = p_demux->p_sys;
 
-    if( p_sys->i_pmt_es <= 0 )
-        return;
+    block_t     *p_pkt;
+
+    /* Get a new TS packet */
+    if( !( p_pkt = stream_Block( p_demux->s, p_sys->i_packet_size ) ) )
+    {
+        msg_Dbg( p_demux, "eof ?" );
+        return NULL;
+    }
+
+    /* Check sync byte and re-sync if needed */
+    if( p_pkt->p_buffer[0] != 0x47 )
+    {
+        msg_Warn( p_demux, "lost synchro" );
+        block_Release( p_pkt );
+        while( vlc_object_alive (p_demux) )
+        {
+            const uint8_t *p_peek;
+            int i_peek, i_skip = 0;
+
+            i_peek = stream_Peek( p_demux->s, &p_peek,
+                    p_sys->i_packet_size * 10 );
+            if( i_peek < p_sys->i_packet_size + 1 )
+            {
+                msg_Dbg( p_demux, "eof ?" );
+                return NULL;
+            }
+
+            while( i_skip < i_peek - p_sys->i_packet_size )
+            {
+                if( p_peek[i_skip] == 0x47 &&
+                        p_peek[i_skip + p_sys->i_packet_size] == 0x47 )
+                {
+                    break;
+                }
+                i_skip++;
+            }
+            msg_Dbg( p_demux, "skipping %d bytes of garbage", i_skip );
+            stream_Read( p_demux->s, NULL, i_skip );
+
+            if( i_skip < i_peek - p_sys->i_packet_size )
+            {
+                break;
+            }
+        }
+        if( !( p_pkt = stream_Block( p_demux->s, p_sys->i_packet_size ) ) )
+        {
+            msg_Dbg( p_demux, "eof ?" );
+            return NULL;
+        }
+    }
+    return p_pkt;
+}
+
+static mtime_t GetPCR( block_t *p_pkt )
+{
+    const uint8_t *p = p_pkt->p_buffer;
+
+    mtime_t i_pcr = -1;
 
     if( ( p[3]&0x20 ) && /* adaptation */
         ( p[5]&0x10 ) &&
         ( p[4] >= 7 ) )
     {
         /* PCR is 33 bits */
-        const mtime_t i_pcr = ( (mtime_t)p[6] << 25 ) |
-                              ( (mtime_t)p[7] << 17 ) |
-                              ( (mtime_t)p[8] << 9 ) |
-                              ( (mtime_t)p[9] << 1 ) |
-                              ( (mtime_t)p[10] >> 7 );
+        i_pcr = ( (mtime_t)p[6] << 25 ) |
+                ( (mtime_t)p[7] << 17 ) |
+                ( (mtime_t)p[8] << 9 ) |
+                ( (mtime_t)p[9] << 1 ) |
+                ( (mtime_t)p[10] >> 7 );
+    }
+    return i_pcr;
+}
+
+static int SeekToPCR( demux_t *p_demux, int64_t i_pos )
+{
+    demux_sys_t *p_sys = p_demux->p_sys;
+
+    mtime_t i_pcr = -1;
+
+    int64_t i_initial_pos = stream_Tell( p_demux->s );
+
+    if( i_pos < 0 )
+        return VLC_EGENERIC;
+
+    int64_t i_last_pos = i_pos + p_sys->i_packet_size * 4500;
+    if( i_last_pos > stream_Size( p_demux->s ) - p_sys->i_packet_size )
+    {
+        i_last_pos = stream_Size( p_demux->s ) - p_sys->i_packet_size;
+    }
+
+    if( stream_Seek( p_demux->s, i_pos ) )
+        return VLC_EGENERIC;
+
+    while( vlc_object_alive( p_demux ) )
+    {
+        block_t     *p_pkt;
+        if( !(p_pkt = ReadTSPacket( p_demux )) )
+        {
+            break;
+        }
+
+        if( p_sys->i_pmt_es > 0 )
+        {
+            if( PIDGet( p_pkt ) == p_sys->i_pid_first_pcr )
+            {
+                i_pcr = GetPCR( p_pkt );
+            }
+        }
+        block_Release( p_pkt );
+
+        if( i_pcr >= 0 )
+            break;
+
+        if( stream_Tell( p_demux->s ) >= i_last_pos )
+            break;
+    }
+    if( i_pcr < 0 )
+    {
+        stream_Seek( p_demux->s, i_initial_pos );
+        return VLC_EGENERIC;
+    }
+    else
+    {
+        p_sys->i_current_pcr = i_pcr;
+        return VLC_SUCCESS;
+    }
+}
+
+static int Seek( demux_t *p_demux, double f_percent )
+{
+    demux_sys_t *p_sys = p_demux->p_sys;
+
+    int64_t i_initial_pos = stream_Tell( p_demux->s );
+
+    /*
+     * Seek i_msec from current position.
+     * If seeking to over/under range, seek to first, or seek to last - 2sec.
+     * If cannot seek, return VLC_EGENERIC. 
+     */
+    mtime_t i_target_pcr = (p_sys->i_last_pcr - p_sys->i_first_pcr) * f_percent + p_sys->i_first_pcr;
+    int64_t i_msec = (i_target_pcr - p_sys->i_current_pcr) * 100 / 9 / 1000;
+    int64_t i_rate = stream_Size( p_demux->s ) * INT64_C(1000000) / ((p_sys->i_last_pcr - p_sys->i_first_pcr) * 100 / 9);
+    int64_t i_size = stream_Size( p_demux->s );
+
+    /*
+     * Seek by estimated rate until i_msec becomes between -5000 and 5000.
+     * If it does not converge to within 5 loops, treated as a failure.
+     */
+    for( int i = 0; i < 5 && (i_msec < -5000 || 5000 < i_msec); i++ )
+    {
+        int64_t i_pos = stream_Tell( p_demux->s );
+
+        if( i_msec < 0 && i_pos <= i_rate ||
+            i_msec > 0 && i_pos >= i_size - i_rate * 2)
+            return VLC_SUCCESS;
+
+        int64_t i_approximate_pos = i_pos + i_rate * i_msec / 1000;
+        if( i_approximate_pos > i_size - i_rate * 2 )
+            i_approximate_pos = i_size - i_rate * 2;
+        if( i_approximate_pos < 0 )
+            i_approximate_pos = 0;
+
+        mtime_t i_prev_pcr = p_sys->i_current_pcr;
+        if( SeekToPCR( p_demux, i_approximate_pos ) )
+            break;
+
+        /*
+         * Adjusting i_msec.
+         */
+        if( p_sys->i_current_pcr - i_prev_pcr < 0 )
+        {
+            i_msec += (i_prev_pcr - p_sys->i_current_pcr) * 100 / 9 / 1000;
+        }
+        else
+        {
+            i_msec -= (p_sys->i_current_pcr - i_prev_pcr) * 100 / 9 / 1000;
+        }
+    }
+    if( i_msec < -5000 || 5000 < i_msec )
+    {
+        msg_Dbg( p_demux, "cannot find a seek position" );
+        stream_Seek( p_demux->s, i_initial_pos );
+        return VLC_EGENERIC;
+    }
+
+    if( i_msec < 0 )
+    {
+        /*
+         * Seek before per about a second.
+         * If it does not converge to within 10 loops, treated as a failure.
+         */
+        for( int i = 0; i < 10 && i_msec < -500; i++ )
+        {
+            int64_t i_pos = stream_Tell( p_demux->s );
+            if(i_pos <= i_rate )
+                return VLC_SUCCESS;
+
+            mtime_t i_prev_pcr = p_sys->i_current_pcr;
+            if( SeekToPCR( p_demux, i_pos - i_rate ) )
+                break;
+            if( i_prev_pcr - p_sys->i_current_pcr < 0 )
+                break;
+            i_msec += (i_prev_pcr - p_sys->i_current_pcr) * 100 / 9 / 1000;
+        }
+        if( i_msec < -500 )
+        {
+            msg_Dbg( p_demux, "cannot find a seek position" );
+            stream_Seek( p_demux->s, i_initial_pos );
+            return VLC_EGENERIC;
+        }
+    }
+    else
+    {
+        /*
+         * Seek after per about a second.
+         * If it does not converge to within 10 loops, treated as a failure.
+         */
+        for( int i = 0; i < 10 && i_msec > 500; i++ )
+        {
+            int64_t i_pos = stream_Tell( p_demux->s );
+            if( i_pos >= i_size - i_rate * 2 )
+                return VLC_SUCCESS;
+
+            mtime_t i_prev_pcr = p_sys->i_current_pcr;
+            if( SeekToPCR( p_demux, i_pos + i_rate ) )
+                break;
+            if( p_sys->i_current_pcr - i_prev_pcr < 0 )
+                break;
+            i_msec -= (p_sys->i_current_pcr - i_prev_pcr) * 100 / 9 / 1000;
+        }
+        if( i_msec > 500 )
+        {
+            msg_Dbg( p_demux, "cannot find a seek position" );
+            stream_Seek( p_demux->s, i_initial_pos );
+            return VLC_EGENERIC;
+        }
+    }
+    return VLC_SUCCESS;
+}
+
+static void GetFirstPCR( demux_t *p_demux )
+{
+    demux_sys_t *p_sys = p_demux->p_sys;
+
+    p_sys->i_first_pcr = -1;
+
+    int64_t i_initial_pos = stream_Tell( p_demux->s );
+
+    while( vlc_object_alive (p_demux) )
+    {
+        block_t     *p_pkt;
+        if( !(p_pkt = ReadTSPacket( p_demux )) )
+        {
+            break;
+        }
+
+        /* Parse the TS packet */
+        ts_pid_t *p_pid = &p_sys->pid[PIDGet( p_pkt )];
+
+        if( p_pid->b_valid )
+        {
+            if( p_pid->psi )
+            {
+                if( p_pid->i_pid == 0 || ( p_sys->b_dvb_meta && ( p_pid->i_pid == 0x11 || p_pid->i_pid == 0x12 || p_pid->i_pid == 0x14 ) ) )
+                {
+                    dvbpsi_PushPacket( p_pid->psi->handle, p_pkt->p_buffer );
+                }
+                else
+                {
+                    for( int i_prg = 0; i_prg < p_pid->psi->i_prg; i_prg++ )
+                    {
+                        dvbpsi_PushPacket( p_pid->psi->prg[i_prg]->handle,
+                                           p_pkt->p_buffer );
+                    }
+                }
+            }
+        }
+        if( p_sys->i_pmt_es > 0 )
+        {
+            mtime_t i_pcr = GetPCR( p_pkt );
+            if ( i_pcr >= 0 )
+            {
+                p_sys->i_pid_first_pcr = p_pid->i_pid;
+                p_sys->i_first_pcr = i_pcr;
+            }
+        }
+        block_Release( p_pkt );
+        
+        if( p_sys->i_first_pcr >= 0 )
+            break;
+    }
+
+    stream_Seek( p_demux->s, i_initial_pos );
+}
+
+static void GetLastPCR( demux_t *p_demux )
+{
+    demux_sys_t *p_sys = p_demux->p_sys;
+
+    p_sys->i_last_pcr = -1;
+
+    int64_t i_initial_pos = stream_Tell( p_demux->s );
+
+    int64_t i_last_pos = stream_Size( p_demux->s ) - p_sys->i_packet_size;
+    int64_t i_pos = i_last_pos - p_sys->i_packet_size * 4500; /* FIXME if the value (4500) is not reasonable, please change it. */
+    if( i_pos < 0 )
+        return;
+
+    if( stream_Seek( p_demux->s, i_pos ) )
+        return;
+
+    while( vlc_object_alive( p_demux ) )
+    {
+        block_t     *p_pkt;
+        if( !(p_pkt = ReadTSPacket( p_demux )) )
+        {
+            break;
+        }
+
+        if( p_sys->i_pmt_es > 0 )
+        {
+            if( PIDGet( p_pkt ) == p_sys->i_pid_first_pcr )
+            {
+                mtime_t i_pcr = GetPCR( p_pkt );
+                if( i_pcr >= 0 )
+                {
+                    p_sys->i_last_pcr = i_pcr;
+                }
+            }
+        }
+        block_Release( p_pkt );
+        if( stream_Tell( p_demux->s ) >= i_last_pos )
+            break;
+    }
+
+    stream_Seek( p_demux->s, i_initial_pos );
+}
+
+static void PCRHandle( demux_t *p_demux, ts_pid_t *pid, block_t *p_bk )
+{
+    demux_sys_t   *p_sys = p_demux->p_sys;
+
+    if( p_sys->i_pmt_es <= 0 )
+        return;
+
+    mtime_t i_pcr = GetPCR( p_bk );
+    if( i_pcr >= 0 )
+    {
+        if( p_sys->i_pid_first_pcr == pid->i_pid )
+        {
+            p_sys->i_current_pcr = i_pcr;
+        }
 
         /* Search program and set the PCR */
         for( int i = 0; i < p_sys->i_pmt; i++ )
-- 
1.7.4.4

